///////////////////////////////////////////////////////////////
//  Gaussel.cpp  -  Gaussian Elimination                     //
//  ver 1.2         only for 3 by 4 augmented matrix         //
//                                                           //
//  Language:       Visual C++, ver 6.0,  used like C        //
//  Platform:       Micron Dual Pentium Pro 200, win NT 4.0  //
//  Application:    Prototype for CSE691, project #2         //
//  Author:         Jim Fawcett, CST 2-187, x3948            //
//                  fawcett@ecs.syr.edu                      //
///////////////////////////////////////////////////////////////

#include "gaussel.h"
using namespace std;

//----< test assertion and, if false, abort with message >-----
//
//  inputs:
//    - test:   an expression which evaluates to true or false
//    - s:      a string displayed in a message box if test is false
//  outputs:
//    - message box if test is false, otherwise nothing
//
static void verify(bool test, char *s) {

  if(!test) {
	  cout << "\n  " << s << endl;
    throw(s);
  }
}

//----< create 3 by 4 matrix >---------------------------------
//
//  outputs:
//    - returns a pointer to a 3 by 4 matrix structure if successful;
//      otherwise program terminates with a diagnostic message
//
long double** gaussel::createMat() {

  long double **m = new long double*[3];
  verify(m!=NULL,"memory alloc failure in createMat()");
  int i;
  for(i=0; i<3; i++) {
    m[i] = new long double[4];
    verify(m[i]!=NULL,"memory alloc failure in createMat()");
  }
  return m;
}

//----< destroy matrix >---------------------------------------
//
//  inputs:
//    - mat:    pointer to valid 3 by 4 matrix structure
//  outputs:
//    -  mat:   NULL pointer
//
void gaussel::destroyMat(long double **mat) {
  int i;
  for(i=0; i<3; i++)
    delete [] mat[i];
  delete [] mat;
  mat = NULL;
}

//----< interchange row pointers >-----------------------------
//
//  inputs:
//    - amat:   pointer to a 3 by 4 matrix structure
//    - row1:   integer specifying one of the rows to swap
//    - row2:   integer specifying one of the rows to swap
//  outputs:
//    - amat:   points to mat with swapped rows
//
static void swapRows(long double **amat, int row1, int row2) {

  verify(0<=row1 && row1 <=2,"row index error in swapRows");
  verify(0<=row2 && row2 <=2,"row index error in swapRows");
  long double *temp = amat[row1];
  amat[row1] = amat[row2];
  amat[row2] = temp;
}

//----< multiply a row by scalar >-----------------------------
//
//  inputs:
//    - amat:   pointer to a 3 by 4 matrix structure
//    - row:    index of row to be modified
//    - mult:   factor used to multiply each element of row
//  outputs:
//    - amat:   points to mat with modified row
//
static void mulRow(long double **amat, int row, long double mult) {

  verify(0<=row && row <=2,"row index error in mulRow");
  int i;
  for(i=0; i<4; i++)
    amat[row][i] *= mult;
}

//

//----< replace row2 by row2 - factor*row1 >-------------------
//
//  inputs:
//    - amat:   pointer to 3 by 4 matrix structure
//    - row1:   index of row to multiply by factor before
//              subtracting from row2
//    - row2:   only row modified by elimination
//    - factor: holds first element of row2 divided by
//              first element of row1
//  outputs:
//    - amat:   points to mat with zeros before diagonal in row2
//
static void 
elim(long double **amat, int row1, int row2, long double factor) {

  verify(0<=row1 && row1 <=2,"row index error in elim(...)");
  verify(0<=row2 && row2 <=2,"row index error in elim(...)");
  int i;
  for(i=0; i<4; i++)
    amat[row2][i] -= factor*amat[row1][i];
}

//----< move largest element to diagonal by swapping rows >----
//
//  inputs:
//    - amat:   pointer to 3 by 4 matrix structure
//    - col:    index of column holding pivot element.  Pivot
//              lies on diagonal
//  outputs:
//    - amat:   points to matrix with pivot at least as large as
//              any element below it in same column
//
static void pivot(long double **amat, int col) {

  verify(0<=col && col <=3,"col index error in pivot(...)");
  long double val = 0;
  if(col==0) {
    if(fabs(amat[0][0]) < fabs(amat[1][0]))
      swapRows(amat,0,1);
    if(fabs(amat[0][0]) < fabs(amat[2][0]))
      swapRows(amat,0,2);
  }
  if(col==1)
    if(fabs(amat[1][1]) < fabs(amat[2][1]))
      swapRows(amat,1,2);
}

//

//----< reduce to diagonal form by Gauss elimination >---------
//
//  inputs:
//    - amat:   pointer to 3 by 4 matrix structure
//  outputs:
//    - amat:   points to matrix in lower echelon form
//
void gaussel::GaussEl(long double **amat) {

  pivot(amat,0);
  elim(amat,0,1,amat[1][0]/amat[0][0]);
  elim(amat,0,2,amat[2][0]/amat[0][0]);
  pivot(amat,1);
  elim(amat,1,2,amat[2][1]/amat[1][1]);
}

//----< find solution by back substitution >-------------------
//
//  inputs:
//    - amat:   pointer to 3 by 4 matrix structure
//  outputs:
//    - amat:   points to matrix with solution in last column
//
void gaussel::backSub(long double **amat) {

  amat[2][3] = amat[2][3]/amat[2][2];
  amat[1][3] = (amat[1][3] - amat[1][2]*amat[2][3])/amat[1][1];
  amat[0][3] = 
    (amat[0][3] - amat[0][1]*amat[1][3] - amat[0][2]*amat[2][3])/amat[0][0];
}

//----< extract solution from augmented matrix >---------------
//
//  inputs:
//    - amat:   pointer to 3 by 4 matrix structure
//    - vec:    pointer to array of 3 doubles
//  outputs:
//    - vec:    holds solution of system of 3 equations in 3 unknowns
//
void gaussel::sol(long double **amat, long double *vec) {

  vec[0] = amat[0][3];
  vec[1] = amat[1][3];
  vec[2] = amat[2][3];
}

//

//----< multiply matrix by solution >--------------------------
//
//  inputs:
//    - amat:   pointer to 3 by 4 matrix structure
//    - vec:    pointer to array of 3 doubles holding contents
//              of solution vector
//  outputs:
//    - amat:   points to matrix with product of amat*vec in last column
//
static void check(long double **amat, long double *vec) {

  amat[0][3] = amat[0][0]*vec[0] + amat[0][1]*vec[1] + amat[0][2]*vec[2];
  amat[1][3] = amat[1][0]*vec[0] + amat[1][1]*vec[1] + amat[1][2]*vec[2];
  amat[2][3] = amat[2][0]*vec[0] + amat[2][1]*vec[1] + amat[2][2]*vec[2];
}



//----< send augmented matrix to standard out >----------------
//
//  inputs:
//    - s:      message to be displayed
//    - amat:   pointer to 3 by 4 matrix structure
//  outputs:
//    - message and matrix sent to standard output
//
void gaussel::display(char *s, long double **amat) {

  cout << s << "\n\n" << 3 << ' ' << 4 << "\n\n";
  long save = cout.flags();
  cout.setf(ios::showpoint);
  cout.setf(ios::fixed,ios::floatfield);
  cout.precision(4);
  int i,j;
  for(i=0; i<3; i++) {
    for(j=0; j<4; j++) {
      cout.width(12);
      cout << amat[i][j];
    }
    cout << endl;
  }
  cout << endl;
  cout.flags(save);
}

gaussel::gaussel()
{}

gaussel::~gaussel()
{}

//--------------------< test stub >---------------------------

#ifdef TEST_GAUSSEL

void main() {

  long double amat[3][4] = {
    1, 2, 3, 1,
    2, 3, 4, 1,
    3, 8, 3, 1
  };

  long double **m = createMat();
  copy(amat,m);

  long double vec[3];

  display("initial augmented matrix",m);
  GaussEl(m);
  display("after Gauss elimination",m);
  backSub(m);
  display("after backsubstitution",m);
  sol(m,vec);
  copy(amat,m);
  check(m,vec);
  display("checking solution",m);
  destroyMat(m);
}

#endif
